## EVIDENCEB, 2020
## create marks for recommendation algorithm
## File description:
## create_marks function

from LrnXPAnaToolbox.lib import *
import pandas as pds
import numpy as np
import os
import sys
from tqdm import tqdm
from tqdm.auto import tqdm
import datetime
from pandas import Panel

def usage(argv): # argv = arguments passed to the script = create_marks.py -h
    if len(argv) == 2:
        if argv[1] == "-h":
            print("\nUSAGE")
            print("    ./create_marks.py  pickle_file")
            print("\nDESCRIPTION")
            print("""    pickle_file       Name of the pickle file to use as database, if the file is called chatbot_data.pk1, please enter chatbot_data as pickle_file.""")
            print("""\ncreate_marks.py must be into a directory with a directory named \'data\' with the pickle file :\nchatbot_data.pk1""")
            return 1
    return 0

def check_error(argv): # argv = arguments passed to the script = create_marks.py pickle_file
    if len(argv) != 2:
        print("Wrong number of arguments, try \"-h\" for more informations")
        return 1
    if (argv[1].isidentifier() != True)|(NameError) :
        print("Argument have to be the name of a pickle file so : alphanumeric letters (a-z) and/or (0-9) and/or underscores (_).")
        return 1
    return 0

def conditionnal_proba(dataset, module, path, test, failure=False, get_proba_specific_path=False):
    """ conditionnal proba function returns a dictionnary with a couple of question as key :
        (q1, q2) and 2 elements as values :
        - the first one : a contingency table
        - the second one : the probability value of the conditionnal event : answer correctly to q2 given q1 correct
        This function has 2 boolean parameters set as default to False :
        - failure which specifies if we want to get the conditionnal probability of success (if false) or failure (if true)
        - get_proba_specific_path specifies if we want to specify a path and a module to focus on (if true) or not (if false)
        module and path are unuseful parameters if get_proba_specific_path is set as False so just put an integer to launch
        the function ; if get_proba_specific_path is set as True, module and path are here to specify the path to focus on and
        from which we want to get the probability matrix of questions.

        EXAMPLE ------------------------------------------------------
        import pandas as pds
        import numpy as np
        import os
        database = pds.read_pickle(os.path.join(".","data", "chatbot_data.pk1")).iloc[:5]
        dic_success = conditionnal_proba(database, module=-1, path=-1, test=1, failure=False, get_proba_specific_path=False)
        # one element of the dictionnary : ( (2, 0, 0, 6),(2, 0, 0, 5) ) : array( [ array([[0., 0.],[0., 1.]]), 1.0 ], dtype=object ) """

    if get_proba_specific_path == True :
        data = dataset[(dataset['module']==module)&(dataset['path']==path)&(dataset['essai']==test)&(dataset['etape']!=3)].loc[:,['id_eleve','id_mpae','correct','essai']]
    else :
        data = dataset[(dataset['etape']!=3)&(dataset['essai']==test)][['id_eleve','id_mpae','correct','essai']]
    dic_data = {}
    for i in range(data.shape[0]):
        # for each student as dictionnary key we give a list [id_mpae, correct, test number]
        key = data.iloc[i]['id_eleve']
        if key not in dic_data:
            dic_data[key] = []
        dic_data[key].append(tuple(data.iloc[i]))
    P = {}
    for key, values in dic_data.items():
        for i, vi in enumerate(values):
            for j, vj in enumerate(values):
                if i == j:
                    continue
                q1 = vi[1:-1]
                q2 = vj[1:-1]
                couple = q1[0], q2[0]
                if couple not in P:
                    P[couple] = np.array([ np.zeros((2, 2)) , 0 ])
                # to complete the contingency table :
                # if q1 correct : q1_ok = 1 else q1 uncorrect : q1_ok = 0 :
                q1_ok = 1 if q1[-1] else 0
                # if q2 correct : q2_ok = 1 else q2 uncorrect : q2_ok = 0 :
                q2_ok = 1 if q2[-1] else 0
                # [q1_ok,q2_ok] is the box in the grid where we count the number of participants
                P[couple][0][q1_ok, q2_ok] += 1
    for couple in P :
        # reminder : to calcuate the conditionnal probability of column given row we do :
        # probability of column & row / probability of row
        # to compute the probability value :
        if failure == True:
            proba_ij = P[couple][0][0,0]
            proba_i = P[couple][0][0,0] + P[couple][0][0,1]
        else:
            proba_ij = P[couple][0][1,1]
            proba_i = P[couple][0][1,0] + P[couple][0][1,1]
        # reminder : P(answer correctly to q2 given q1 correct ) = P(q1 is correct & q2 is correct) / P(q1 is correct)
        if proba_i != 0 :
            P[couple][1] += proba_ij/proba_i
    return P

def complete_proba_matrix(proba, failure=False):
    """ complete_proba_matrix return 2 dataframe (= table = matrix). The first one representing the conditionnal
        probability matrix of success or failure for answering to question in column given the answer to question
        in row. The second one with the value of the participant for each couple of question. The proba parameter
        is a dictionnary of the form of the one returned by conditionnal_proba. The failure parameter is the same
        as previously : it specifies if we want to get the conditionnal probability of success (if false) or
        failure (if true).

        EXAMPLE ------------------------------------------------------
        import pandas as pds
        import numpy as np
        import os
        database = pds.read_pickle(os.path.join(".","data", "chatbot_data.pk1")).iloc[:5]
        dic_success = conditionnal_proba(database, module=-1, path=-1, test=1, failure=False, get_proba_specific_path=False)
        # one element of the dictionnary : ( (2, 0, 0, 6),(2, 0, 0, 5) ) : array( [ array([[0., 0.],[0., 1.]]), 1.0 ], dtype=object )
        success_matrix, participants_success = complete_proba_matrix(proba=dic_success, failure=False)
        # one row of the pandas dataframe called success_matrix is :
        #           row            col         value
        #   7   (2, 0, 0, 6)   (2, 0, 0, 5)     1.0 """

    couple_list_0 = np.sort(pds.Series([couple[0] for couple in list(proba.keys())]).unique())
    couple_list_1 = np.sort(pds.Series([couple[1] for couple in list(proba.keys())]).unique())
    matrix=[]
    participants = []
    for i,couple in enumerate(proba):
        v = proba[couple]
        matrix.append(dict(row=couple[0],col=couple[1],value=v[1]))
        if failure == True:
            participants.append(dict(row=couple[0],col=couple[1],value=v[0][0,0] + v[0][0,1]))
        else:
            participants.append(dict(row=couple[0],col=couple[1],value=v[0][1,0] + v[0][1,1]))
    return pds.DataFrame(matrix), pds.DataFrame(participants)

def create_marks(pickle_file) :

    """ create_marks returns a dataframe with 3 columns the student, the question and the mark. This mark used to represent
        the learning achieved by the student with the exercise. More the mark has a high value, more the question is useful
        for the learning of the student. One of the tenets of this function is that : making mistakes is useful.
        This function takes 1 parameters :
        - the pickle file (at the string format) to transform to pandas dataframe which used to be of the form :
          (You can get one pickle like this one with LrnXPAnaToolbox.transform_data from a json file).
          columns :
            Index(['id_eleve', 'module', 'path', 'activity', 'tps_posix', 'exercice', 'correct', 'jour', 'etape', 'id_mpa',
            'id_mpae', 'date', 'essai', 'tps+1', 'duree', 'num_connection', 'reconnections', 'nbr_questions_faites',
            'parcours_diagnostique', 'horaire'], dtype='object')
          columns type :
            id_eleve                  object
            module                     int64
            path                       int64
            activity                   int64
            tps_posix                float64
            exercice                   int64
            correct                     bool
            jour                       int64
            etape                      int64
            id_mpa                    object
            id_mpae                   object
            date                      object
            essai                    float64
            tps+1                    float64
            duree                    float64
            num_connection             int64
            reconnections              int64
            nbr_questions_faites       int64
            parcours_diagnostique    float64
            horaire                  float64

        EXAMPLE --------------------------------------------------------------------------------------------------
        import pandas as pds
        import numpy as np
        import os
        from sklearn.feature_extraction import DictVectorizer
        from sklearn.model_selection import train_test_split, GridSearchCV
        from sklearn.tree import DecisionTreeClassifier

        create_marks_2( "chatbot_data.pk1" )

        # one row of the pandas dataframe called returned is :
        #           student                question         mark
        #   97   7085AA80-B1EA-4D...     (2, 0, 0, 5)        3 """

    print("start computing data to create marks")
    print("                        / `.   .' \\")
    database = pds.read_pickle(os.path.join(".","data", f"{pickle_file}.pk1"))
    DF = database[database['etape']!=3]

    from_list_to_str(DF,'id_mpae')

    print("                .---.  <    > <    >  .---.")
    dic_success = conditionnal_proba(DF, module=-1, path=-1, test=1, failure=False, get_proba_specific_path=False)
    print("                |    \  \ - ~ ~ - /  /    |")
    success_matrix, participants_success = complete_proba_matrix(proba=dic_success, failure=False)
    success_matrix[['row','col']].astype(str)
    print("                 ~-..-~             ~-..-~")
    data_success = success_matrix.pivot_table(index='row', columns='col', values='value')
    data_success.apply(pds.to_numeric)
    data_success = data_success.fillna(1)
    print("             \~~~\.'                    `./~~~/")
    dic_failure = conditionnal_proba(DF, module=-1, path=-1, test=1, failure=True, get_proba_specific_path=False)
    print("   .-~~^-.    \__/                        \__/")
    failure_matrix, participants_failure = complete_proba_matrix(proba=dic_failure, failure=True)
    failure_matrix[['row','col']].astype(str)
    print(" .'  O    \     /               /       \  \\")
    data_failure = failure_matrix.pivot_table(index='row', columns='col', values='value')
    data_failure.apply(pds.to_numeric)
    data_failure = data_failure.fillna(1)
    print("(_____,    `._.'               |         }  \/~~~/")
    print(" `----.          /       }     |        /    \__/")
    pos_effect = {question:data_success[question].where(data_success[question]>0.8).index[data_success[question].where(data_success[question]>0.8)==1] for question in DF['id_mpae'].unique()}
    print("       `-.      |       /      |       /      `. ,~~|")
    neg_effect = {question:data_failure.T[question].where(data_failure.T[question]>0.8).index[data_failure.T[question].where(data_failure.T[question]>0.8)==1] for question in DF['id_mpae'].unique()}
    print("           ~-.__|      /_ - ~ ^|      /- _      `..-'   UQDKR")
    q2_time = {question:DF[ (DF.id_mpae==question) & (DF.duree<2700) ]['duree'].describe()['50%'] for question in DF['id_mpae'].unique()}
    print("                |     /        |     /     ~-.     `-. _||_||_")
    q3_time = {question:DF[ (DF.id_mpae==question) & (DF.duree<2700) ]['duree'].describe()['75%'] for question in DF['id_mpae'].unique()}
    print("                |_____|        |_____|         ~ - . _ _ _ _ _>")
    students = np.array([])
    questions = np.array([])
    marks = np.array([])
    dataset = DF
    print('start creating marks')
    groupby_student_question = dataset.groupby(['id_eleve','id_mpae'])

    def give_mark(data_student_question):
        id_eleve, id_mpae = data_student_question[['id_eleve','id_mpae']].iloc[0]
        mark = 1
        question_first_test_date = min(data_student_question.tps_posix)
        question_first_test_connection = data_student_question[ data_student_question.tps_posix==question_first_test_date ]['num_connection'].values[0]
        # We look at the first and second test on the first connection for this question :
        question_first_test = data_student_question[ (data_student_question.essai == 1) & (data_student_question.num_connection == question_first_test_connection) ]
        question_second_test = data_student_question[ (data_student_question.essai == 2) & (data_student_question.num_connection == question_first_test_connection) ]
        if question_first_test.shape[0] != 0:
            if question_first_test.shape[0] != 1:
                print(f"ERROR question_first_test must designate the first test done : it must be a 1 line table but it's not the case : {question_first_test.shape}")
            # check the probability for the student to dropout :
            student_tried_idmpae = data_student_question[ data_student_question.essai==1 ]
            question_connections = np.array(student_tried_idmpae['num_connection'].unique())
            question_dates = np.array(student_tried_idmpae['tps_posix'].unique())
            # check if the student didnt try any other exercise after the question
            df_student_other_question = dataset[ (dataset.id_eleve==id_eleve) & (dataset.id_mpae!=id_mpae) ]
            def student_didnt_try_anything_after(i_connection):
                return df_student_other_question[ (df_student_other_question.num_connection==question_connections[i_connection]) & (df_student_other_question.tps_posix>question_dates[i_connection]) ].shape[0] == 0
            student_didnt_try_anything_after_vectorize = np.vectorize(student_didnt_try_anything_after)
            idx_question_connections = np.arange(len(question_connections))
            # we apply student_didnt_try_anything_after for all the index of the array question_connections
            # to get the array of True False where the condition student_didnt_try_anything_after() is satisfied
            student_didnt_try_anything_after = sum(student_didnt_try_anything_after_vectorize(idx_question_connections))
            probability_of_dropout = student_didnt_try_anything_after / student_tried_idmpae.shape[0]
            if probability_of_dropout > 0.5 :
                mark -= 1
            else :
                mark += 1
            # test if the average reponse time is quite long whereas the student got right at the first try
            if question_first_test['correct'].iloc[0] == True:
                if question_first_test['tps_posix'].iloc[0] > (q3_time[id_mpae]+q2_time[id_mpae]) / 2 :
                    mark += 1
            else:
                # test if the student got right at the second try
                if question_second_test.shape[0] != 0 :
                    if question_second_test['correct'].iloc[0] == True :
                        mark += 1
            # test if the student failed any question that has a negative influence at a previous date and succeeded here (at 1st or 2nd test)
            if dataset[ (dataset.id_eleve == id_eleve) & (dataset.correct == False) & (dataset.tps_posix <= question_first_test_date) ]['id_mpae'].isin(neg_effect).sum() > 1 :
                if question_first_test['correct'].iloc[0] == True :
                    mark += 1
                elif question_second_test.shape[0] != 0 :
                    if question_second_test['correct'].iloc[0] == True :
                        mark += 1
            # test if the student succeeded any question that has a positive influence at a previous date and failed here (at 1st or 2nd test)
            if dataset[ (dataset.id_eleve == id_eleve) & (dataset.correct == True) & (dataset.tps_posix <= question_first_test_date) ]['id_mpae'].isin(pos_effect).sum() > 1 :
                if question_first_test['correct'].iloc[0] == False :
                    mark += 1
                elif question_second_test.shape[0] != 0 :
                    if question_second_test['correct'].iloc[0] == False :
                        mark += 1

        return id_eleve, str_question_tolist(id_mpae), int(mark)

    tqdm.pandas()
    resultat = groupby_student_question.progress_apply(give_mark)
    return pds.DataFrame(list(resultat), columns = ['student', 'question', 'mark'])

def main(argv):
    if usage(argv) != 0:
        exit(0)
    if check_error(argv) != 0:
        exit(84)
    pickle = str(argv[1])
    create_marks(pickle)
    return 0

if __name__ == '__main__':
    main(sys.argv)


"""Top-level package for UQDKR LrnXPAnaToolbox."""

__author__ = """Kayané Elmayan Robach"""
__email__ = 'kaya.robach@gmail.com'
__version__ = '0.3.3'

from .lib import str_question_tolist
from .lib import list_question_tostr
from .lib import from_list_to_str
from .lib import from_str_to_list

from .LrnXPAnaToolbox import test

from .cluster_students.cluster_students import plot_clustering
from .cluster_students.cluster_students import creationDFClustering
from .cluster_students.cluster_students import optimal_n_clusters
from .cluster_students.cluster_students import similar_students

from .create_marks.create_marks import conditionnal_proba
from .create_marks.create_marks import complete_proba_matrix
from .create_marks.create_marks import create_marks

from .transform_data.transform_data import to_workable_json
from .transform_data.transform_data import json_to_csv_seq
from .transform_data.transform_data import csv_to_dataframe
from .transform_data.transform_data import dict_diagnostique
from .transform_data.transform_data import json_to_csv_track
from .transform_data.transform_data import remplir_etape
from .transform_data.transform_data import index_essay
from .transform_data.transform_data import id_mpae_3_step
from .transform_data.transform_data import set_essai
from .transform_data.transform_data import duree
from .transform_data.transform_data import plus_proche
from .transform_data.transform_data import get_index_connection
from .transform_data.transform_data import get_parcours_diagnostique
from .transform_data.transform_data import tranche_horaire
from .transform_data.transform_data import transform_data

from .recommender.dropout import dropout_prediction_training_data
from .recommender.dropout import dropout_prediction
from .recommender.dropout import dropout_recommendation

from .recommender.final_recommendation import recom_algorithm

from .recommender.recommender import algorithm
from .recommender.recommender import get_files



import pandas as pds
from math import sqrt
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import os
from LrnXPAnaToolbox.lib import *

def recom_algorithm(userinput_df, students_df) :
    """Select 20 questions with the best relevance score to present to the student using a collaborative filtering."""
    userinput_df = from_list_to_str(userinput_df,'question_id')
    students_df = from_list_to_str(students_df,'question_id')
    
    # Beginning of the collaborative filtering.
    # Filtering out students that have answered the same questions as the student X and storing it
    usersubset = students_df[students_df['question_id'].isin(userinput_df['question_id'])]
    # Creating sub dataframes for every students
    usersubsetgroup = usersubset.groupby(['student_id'])
    # Sorting it so students with questions most in common with the student X will have priority
    usersubsetgroup = sorted(usersubsetgroup, key=lambda x: len(x[1]), reverse=True)
    # Creating the Pearson Correlation Coefficient dictionnary where the key is the student id and
    # the value is the coefficient
    pearson_co = {}
    # For every student group in our subdataframes
    for name, group in usersubsetgroup:
    # Sort the student X and current student subdataframe so values are't mixed
        group = group.sort_values(by='question_id')
        userinput_df = userinput_df.sort_values(by='question_id')
        # Get 'n' for pearson formula
        n = len(group)
        # Get ratings that both students have in common in a temporary dataframe
        tmp_df = userinput_df[userinput_df['question_id'].isin(group['question_id'].tolist())]
        # Then store it in a temporary dataframe
        tmp_rating = tmp_df['rating'].tolist()
        # Put current student marks subdataframe in a list
        tmp_group = group['rating'].tolist()
        # Now calculate Pearson Correlation between the two students (so called x & y)
        uxx = sum([i**2 for i in tmp_rating]) - pow(sum(tmp_rating), 2) / float(n)
        uyy = sum([i**2 for i in tmp_group]) - pow(sum(tmp_group), 2) / float(n)
        uxy = sum(i*j for i,j in zip(tmp_rating, tmp_group)) - sum(tmp_rating) * sum(tmp_group)/float(n)
        # If the denominator is different than 0, divide, else there is no correlation (0)
        if uxx != 0 and uyy != 0:
            pearson_co[name] = uxy/sqrt(uxx*uyy)
        else:
            pearson_co[name] = 0
    # Convert dictionnary into dataframe, adding column with the similarity index :
    # Value vary from -1 to 1, where 1 is a direct correlation between the students, and -1
    # is a negative correlation
    pearson_df = pds.DataFrame.from_dict(pearson_co, orient='index')
    pearson_df.columns = ['similarity_index']
    pearson_df['student_id'] = pearson_df.index
    pearson_df.index = range(len(pearson_df))
    # Puts top 50 students that are the most similar to student X
    topstuds = pearson_df.sort_values(by='similarity_index', ascending=False)[0:50]

    # Beggining of recommendation
    # Merging similarity index dataframe and question_id dataframe
    topstudsrating = topstuds.merge(students_df, left_on='student_id', right_on='student_id', how='inner')
    # Multiplying question rating by its weight (similarity index) then sum up the new ratings and divide it by
    # the sum of the weights
    topstudsrating['weighted_rating'] = topstudsrating['similarity_index']*topstudsrating['rating']
    # Applying a sum to the top students after grouping it up by students_id
    tmptopstudsrating = topstudsrating.groupby('question_id').sum()[['similarity_index', 'weighted_rating']]
    tmptopstudsrating.columns = ['sum_similarity_index','sum_weighted_rating']
    # Create an empty dataframe and take weighted average
    recommendation_df = pds.DataFrame()
    recommendation_df['weighted_average_recom_score'] = tmptopstudsrating['sum_weighted_rating']/tmptopstudsrating['sum_similarity_index']
    recommendation_df['question_id'] = tmptopstudsrating.index
    recommendation_df = recommendation_df.sort_values(by='weighted_average_recom_score', ascending=False)
    # Get dataframe of only questions id
    students = students_df.question_id.unique()
    questions_df = pds.DataFrame(students)
    questions_df.columns = ['question_id']
    recommendation_df = recommendation_df.reset_index(drop=True)
    # Sort weighted average and see the top 20 questions the algorithm recommended
    recom = recommendation_df.loc[questions_df['question_id'].isin(recommendation_df['question_id'])]
    pds.set_option("display.max_rows", None, "display.max_columns", None)
    return recom.head(20)


## EVIDENCEB, 2020
## transform_data.py
## File description:
## transform data function

import ijson
import json
import csv
import pandas as pds
import matplotlib.pyplot as plt
import numpy as np
import seaborn
import datetime
import scipy
import itertools
import os
import io
import sys
import time
import datetime
from math import *

def usage(argv): # argv = arguments passed to the script = transform_data.py -h
    if len(argv) == 2:
        if argv[1] == "-h":
            print("\nUSAGE")
            print("    ./transform_data.py  option  original_json_sequences_name  original_json_trackings_name")
            print("\nDESCRIPTION")
            print("""    option = 3       Transform data and save 2 pandas dataframes : sequences and chatbot""")
            print("""    option = 1       Transform data and save 1 pandas dataframe : sequences""")
            print("""    option = 2       Transform data and save 1 pandas dataframe : chatbot""")
            print("""    original_json_sequences_name   original_json_trackings_name   are the names of both json files.""")
            print("""\ntransform_data.py must be into a directory with a directory named \'data\' with the files : \noriginal_json_sequences_name.txt and original_json_sequences_name.json\noriginal_json_trackings_name.txt and original_json_trackings_name.json""")
            return 1
    return 0

def check_error(argv): # argv = arguments passed to the script = transform_data.py option seq track
    if len(argv) != 4:
        print("Wrong number of arguments, try \"-h\" for more informations")
        return 1
    if (argv[1].isdigit() != True) :
        print("Arguments have to be numerical characters (1, 2 or 3).")
        return 1
    if argv[1].isdigit() :
        if not int(argv[1]) in [1,2,3] :
            print("Argument must be 3 if you want to save both : sequences and trackings pandas dataframes.")
            print("Argument must be 1 if you want to save the sequences pandas dataframe.")
            print("Argument must be 2 if you want to save the trackings pandas dataframe.")
            return 1
    if (argv[2].isidentifier() != True)|(argv[3].isidentifier() != True):
        print("names arguments have to be the names of the sequences and trackings json files so : alphanumeric letters (a-z) and/or (0-9) and/or underscores (_).")
        return 1
    return 0

def to_workable_json(namefile):
    """Modify the json file (at the text format) to get a workable json file (at the json format) under the name transformed...json."""
    f = io.open((os.path.join(".","data", "{0}.txt".format(namefile))), 'r', encoding='utf-8')
    content = f.read()
    first_elements, last_elements = content[0:2],content[-3:-1]
    check = (first_elements=="[{") & (last_elements=="}]")
    if check :
        print("The json seems to have the correct format : Let's transform the json")
        content = content[content.find('['):]
        element = "["
        to_insert1 = """{ "data" :"""
        to_insert2 = "}"
        idx = content.index(element)
        content = content[:idx] + to_insert1 + content[idx:] + to_insert2
        result = str(content)
        new_content = io.open((os.path.join(".","data", "transformed_{0}.json".format(namefile))), 'w+', encoding='utf-8')
        new_content.write(result)
        new_content.close()
        print("json transformed")
    else:
        print("The json seems to not have the correct format : Let's try to transform the json")
        content = content[content.find('['):]
        element = "["
        to_insert1 = """{ "data" :"""
        to_insert2 = "}"
        idx = content.index(element)
        content = content[:idx] + to_insert1 + content[idx:] + to_insert2
        result = str(content)
        new_content = io.open((os.path.join(".","data", "transformed_{0}.json".format(namefile))), 'w+', encoding='utf-8')
        new_content.write(result)
        new_content.close()
        print("json transformed")

def json_to_csv_seq(csvfile, jsonfile):
    """Take the json modified from sequences file from the previous function to writw a csv file."""
    with io.open((os.path.join(".","data", "{0}.csv".format(csvfile))), 'w', encoding='utf_8') as csvfile:
        spamwriter = csv.writer(csvfile, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        with io.open(os.path.join(".","data", "{0}.json".format(jsonfile)), encoding='utf_8') as f:
            objects = ijson.items(f, '')
            for datum in objects :
                for i in range(len(list(datum.values())[0])):
                    multi_rows = list(datum.values())[0][i]
                    try :
                        module = multi_rows["module"]
                    except KeyError:
                        module = np.nan
                    try:
                        path = multi_rows["path"]
                    except KeyError:
                        path = np.nan
                    try:
                        activity = multi_rows["activity"]
                    except KeyError:
                        activity = np.nan
                    try :
                        token = multi_rows["token"]
                    except KeyError:
                        token = np.nan
                    for sequence in json.loads(multi_rows["sequence"]):
                        try:
                            real_module = sequence['module']
                        except KeyError:
                            real_module = np.nan
                        try:
                            real_path = sequence['path']
                        except KeyError:
                            real_path = np.nan
                        try:
                            real_activity = sequence['activity']
                        except KeyError:
                            real_activity = np.nan
                        try:
                            real_exercice = sequence['exercice']
                        except KeyError:
                            real_exercice = np.nan
                        try:
                            difficultyseq = sequence['difficulty']
                        except KeyError:
                            difficultyseq = np.nan
                        try:
                            type_answer = sequence["type"]
                        except KeyError:
                            type_answer = np.nan
                        try:
                            reponse_eleve = sequence["content_id"]
                        except KeyError:
                            reponse_eleve = np.nan
                        try:
                            reponse_correcte = sequence["correct_answer"]
                        except KeyError:
                            reponse_correcte = np.nan

                        spamwriter.writerow([real_module, real_path, real_activity, real_exercice, difficultyseq, type_answer, module, path, activity, token, reponse_eleve, reponse_correcte])

def csv_to_dataframe(csvfile,columns_names):
    """Transform the csv file from the previous functio into a dataframe."""
    return pds.read_csv((os.path.join(".","data", "{0}.csv".format(csvfile))),names=columns_names)

def dict_diagnostique(dataframe_chatbot):
    """Create a dictionnary with : for each question from the diagnostic test (id_mpae) in key, the path corresponding in value."""
    chatbot_data_difficulty = dataframe_chatbot[dataframe_chatbot['real_path']==0]
    chatbot_data_difficulty = pds.DataFrame.set_index(chatbot_data_difficulty,(np.arange(len(chatbot_data_difficulty))))
    diagnostique = {}

    # we have to transform the id_mpae column in string :
    dfnew = chatbot_data_difficulty[['id_mpae']].copy(deep=True)
    pds.options.mode.chained_assignment = None
    chatbot_data_difficulty['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )

    for idmpae in chatbot_data_difficulty['id_mpae'].unique():
        diagnostique[idmpae]=[]
        df = chatbot_data_difficulty[chatbot_data_difficulty['id_mpae']==idmpae]['difficultyseq']
        for val in df:
            try :
                json.loads(val)
                diagnostique[idmpae].append(val)
                break
            except ValueError:
                continue
            except TypeError:
                continue
    cpt=0
    for key in diagnostique:
        if not diagnostique[key]:
            cpt+=1
            print(f"None of the difficulty dictionnary is readable with json.loads for the exercice {key}, you must code a solution to transform the string into a workable dictionnary.")
    if cpt==0:
        for key in diagnostique:
            diagnostique[key]=diagnostique[key][0]
    return diagnostique

def json_to_csv_track(csvfile, jsonfile):
    """Take the json modified from trackings file from the previous function to writw a csv file."""
    with io.open((os.path.join(".","data", "{0}.csv".format(csvfile))), 'w') as csvfile:
        spamwriter = csv.writer(csvfile, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        with io.open((os.path.join(".","data", "{0}.json".format(jsonfile)))) as f:
            objects = ijson.items(f, '')
            for datum in objects :
                for i in range(len(list(datum.values())[0])):
                    multi_rows = list(datum.values())[0][i]
                    module = multi_rows["module"]
                    path = multi_rows["path"]
                    activity = multi_rows["activity"]
                    token = multi_rows["token"]
                    for tracking in json.loads(multi_rows["tracking"]):
                        try:
                            tms = tracking["tms"]
                        except KeyError:
                            tms = np.nan
                        try:
                            exerciceId = tracking["exerciceId"]
                        except KeyError:
                            exerciceId = np.nan
                        try:
                            isCorrect = tracking["isCorrect"]
                        except KeyError:
                            isCorrect = np.nan
                        try:
                            question_type = tracking["type"]
                        except KeyError:
                            question_type = np.nan

                        spamwriter.writerow([token, module, path, activity, tms, exerciceId, isCorrect, question_type])
                        

def remplir_etape():
    """L1, L2, L3 are the list of the questions from each of the different steps : 1 for the diagnostic test, 2 for the exercice in the path assigned, 3 for the final test.
       1st step : path and activity = 0
       3rd step : the last activity of each path in each module + 1
       2nd step : other question"""
    # WARNING : must be update for each version of the chatbot
    d={}
    L3=set([(1,1,7),(1,2,6),(1,3,6),(2,1,6),(2,2,5),(2,3,5),(3,1,6),(3,2,4),(3,3,6),(4,1,6),(4,2,4),(4,3,6),(5,1,6),(5,2,6),(5,3,5),(6,1,5),(6,2,5),(6,3,4),(7,1,6),(7,2,7),(7,3,6),(8,1,6),(8,2,4),(8,3,5),(9,1,4),(9,2,5),(9,3,5)])
    for triple in L3:
        d[triple]=3
    L1=set([(1,0,0),(2,0,0),(3,0,0),(4,0,0),(5,0,0),(6,0,0),(7,0,0),(8,0,0),(9,0,0)])
    for triple in L1:
        d[triple]=1
    L2 = set(list(itertools.product(*[[1,2,3,4,5,6,7,8,9],[0,1,2,3],[0,1,2,3,4,5,6,7,]])))
    for triple in L2-L1-L3 :
        d[triple]=2
    return d

def index_essay(df):
    colonne_to_sort = 'tms'
    df = df.sort_values(colonne_to_sort, axis = 0, ascending = True)
    df = pds.DataFrame.set_index(df,(np.arange(len(df))+1))
    df['essai'] = df.index
    return df

def id_mpae_3_step(row):
    if row.etape==3 :
        return [int(-1),int(-1),int(-1),int(-1)]
    else :
        return [int(row.module),int(row.path),int(row.activity),int(row.exerciceId)]

def set_essai(data):
    """For each student and question (id_mpae) we have to number the test done (according to the date).
       For each student and question we isolate a dataframe ; then we sort it by ascending date (tms) and we attribute the index as the test number."""
    # we have to transform the id_mpae column in string :
    dfnew = data[['id_mpae']].copy(deep=True)
    pds.options.mode.chained_assignment = None
    data['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )

    def create_frames(data):
        frames = []
        data_eleve_1_2 = data[(data['etape']!=3)].groupby(['token','id_mpae'])
        data_eleve_3 = data[(data['etape']==3)]
        indexed_data_1_2 = data_eleve_1_2.apply(index_essay)
        #indexed_data_3 = data_eleve_3.apply(index_essay)
        frames.append(indexed_data_1_2)
        frames.append(data_eleve_3)
        return frames

    result = pds.concat( create_frames(data) )
    result = pds.DataFrame.set_index(result,np.arange(result.shape[0]))

    result['id_mpae'] = result.apply(lambda row : id_mpae_3_step(row), axis=1)
    return result

def duree(data):
    """For each student, we look at the date of each line of exercice and we calculate the duration according to the next date.
       The last duration cannot be calculated because there is no next date."""
    frames = []
    for eleve in data['token'].unique():
        data_eleve = data[data['token']==eleve]
        data_eleve = data_eleve.sort_values('tms', axis = 0, ascending = True)
        data_eleve = pds.DataFrame.set_index(data_eleve,np.arange(data_eleve.shape[0]))
        data_eleve_withoutfirst = data_eleve.drop([0],axis=0)
        data_eleve_withoutfirst = data_eleve_withoutfirst.append(data_eleve.iloc[0])
        data_eleve_withoutfirst = pds.DataFrame.set_index(data_eleve_withoutfirst,np.arange(data_eleve_withoutfirst.shape[0]))
        data_eleve_withoutfirst = data_eleve_withoutfirst.sort_values('tms', axis = 0, ascending = True)
        data_eleve['tps+1'] = data_eleve_withoutfirst['tms']
        data_eleve['duree'] = data_eleve['tps+1']-data_eleve['tms']
        data_eleve = pds.DataFrame.set_index(data_eleve,np.arange(data_eleve.shape[0]))
        data_eleve = data_eleve.drop([data_eleve.shape[0]-1],axis=0)
        frames.append(data_eleve)
    result = pds.concat(frames)
    result = pds.DataFrame.set_index(result,np.arange(result.shape[0]))
    return result

def plus_proche(idx,array):
    new_array = abs(np.array(array) - idx)
    result = np.argmin(new_array)
    if idx - array[result] > 0:
        return result + 1
    else:
        return result

def get_index_connection(data):
    """To get the index connection, we look at the big duration (indicating that the next exercise has been done later : at a next connection).
       We save the index of these lines in the dataframe and we look at other questions.
       If the question is done before a big duration, the question is done during the connection ending with this big duration.
       If the question is done later, we are in a new connection."""
    data = data.sort_values('tms', axis=0, ascending=True)
    data = pds.DataFrame.set_index(data,np.arange(data.shape[0]))
    eleve_index_grosses_durees = {}
    for idt in data['token'].unique():
        eleve_index_grosses_durees[idt] = data[(data['duree']>=2700)&(data['token']==idt)].index , data[data['token']==idt].index
    index_connection = {idx:1 for idx in range(data.shape[0])}
    for cle in eleve_index_grosses_durees:
        if len(eleve_index_grosses_durees[cle][0])!=0:
            for idx in eleve_index_grosses_durees[cle][1]:
                idx_proche = plus_proche(idx , eleve_index_grosses_durees[cle][0])
                index_connection[idx] = idx_proche + 1
    return index_connection

def get_parcours_diagnostique(dico):
    """We attribute the diagnotic path of any question of the 1st step looking at the dictionnary of difficulty."""
    if type(dico)==str:
        dico = json.loads(dico)
        for key in dico:
            if dico[key]==1:
                key=key.strip()
                return int(key[4])
    else:
        return np.nan

def tranche_horaire(data):
    """For each student and each connection we look at the major hour in each the connection took place."""
    d={}
    data_eleve_connection = data.groupby(['token','num_connection'])
    for serie in data_eleve_connection :
        df = serie[1].sort_values('tms', axis=0, ascending=True)
        duree = np.array(df['duree'])
        array_horaires = np.array(df['tms'])
        for indice in range(len(array_horaires)):
            array_horaires[indice] = int(datetime.datetime.fromtimestamp(array_horaires[indice]).hour)
        d[serie[0]] = np.quantile(array_horaires,0.5)
    return d


def transform_data(option,original_json_sequences_name,original_json_trackings_name):

    today = datetime.date.today()

    if option == 3: #save 2 pandas dataframe : sequences and trackings

        print("building sequences data\n")
        to_workable_json(original_json_sequences_name)
        json_to_csv_seq('jsonfile_to_csv_sequences2',f'transformed_{original_json_sequences_name}')
        print("                                              ,","                                             ,o","                                             :o",sep="\n")
        dumpseq_1 = csv_to_dataframe('jsonfile_to_csv_sequences2',['real_module','real_path','real_activity','real_exercice','difficultyseq', 'type_answer', 'module','path','activity','id_eleve', 'reponse_eleve', 'reponse_correcte'])
        print("                    _....._                  `:o","                  .'       ``-.                \o","                 /  _      _   \                \o",sep="\n")
        dumpseq_1[['real_module','real_path','real_activity','real_exercice']] = dumpseq_1[['real_module','real_path','real_activity','real_exercice']].apply(pds.to_numeric, errors='coerce')
        print("                :  /*\    /*\  :                 ;o","                |  \_/    \_/  :                  ;o","                (       U      /                  ;o",sep="\n")
        dumpseq_1[['real_module','real_path','real_activity','real_exercice']] = dumpseq_1[['real_module','real_path','real_activity','real_exercice']].astype('Int16')
        print("                 \  (\_____/) /                  /o","                  \   UQDKR  (                  /o","                   \         (                ,o:",sep="\n")
        dumpseq_1 = dumpseq_1.fillna(-1)
        print("                   )          \,           .o;o'           ,o'o'o.","                 ./          /\o;o,,,,,;o;o;''         _,-o,-'''-o:o.","  .             ./o./)        \    'o'o'o''         _,-'o,o'         o",sep="\n")
        dumpseq_1.loc[:,'id_mpae'] = dumpseq_1.apply(lambda row: [int(row.module),int(row.path),int(row.activity),int(row.real_exercice)], axis=1)
        print("  o           ./o./ /       .o \.              __,-o o,o'","  \o.       ,/o /  /o/)     | o o'-..____,,-o'o o_o-'","  `o:o...-o,o-' ,o,/ |     \   'o.o_o_o_o,o--''",sep="\n")
        dumpseq_1.loc[:,'id_final_mpae'] = dumpseq_1.apply(lambda row: [int(row.real_module),int(row.real_path),int(row.real_activity),int(row.real_exercice)], axis=1)
        print("  .,  ``o-o'  ,.oo/   'o /\.o`.","  `o`o-....o'o,-'   /o /   \o \.                       ,o..         o","    ``o-o.o--      /o /      \o.o--..          ,,,o-o'o.--o:o:o,,..:o",sep="\n")
        diagnostique = dict_diagnostique(dumpseq_1)
        print("                  (oo(          `--o.o`o---o'o'o,o,-'''        o'o'o","                   \ o\              ``-o-o''''","    ,-o;o           \o \\","    ,-o;o           \o \\",sep="\n")

        # pandas error with list we have to transform the id_mpae column in string :
        dfnew = dumpseq_1[['id_mpae']].copy(deep=True)
        pds.options.mode.chained_assignment = None
        dumpseq_1['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )
        dumpseq_1['diagnostique'] = dumpseq_1['id_mpae'].map(diagnostique)
        dumpseq_1['id_mpae'] = dfnew['id_mpae'].apply( lambda x : list( int(e) for e in x ) )

        print("   /o/               )o )","  (o(               /o /","   \o\.       ...-o'o /",sep="\n")
        dumpseq_1.to_pickle((os.path.join(".","data", f"sequences_data_{today}.pk1")))
        print("    \ o`o`-o'o o,o,--'","       ```o--'''",sep="\n")
        print("\nsequences dataframe built\n")
        print("building trackings data\n")
        to_workable_json(original_json_trackings_name)
        json_to_csv_track('jsonfile_to_csv_trackings2',f'transformed_{original_json_trackings_name}')
        print("                                                       ______","                                                     /    /_-._",sep="\n")
        print("                                                   / /_ ~~o\  :Y","                                                   / : \~x.  ` ')",sep="\n")
        dumptrack_2 = csv_to_dataframe('jsonfile_to_csv_trackings2',['token','module','path','activity','tms','exerciceId','isCorrect','question_type'])
        dico_etapes = remplir_etape()
        print("                                                  /  |  Y< ~-.__j")
        dumptrack_2['tms'] = dumptrack_2['tms']/1000
        dumptrack_2['jour'] = dumptrack_2.apply(lambda row: datetime.datetime.fromtimestamp(row.tms).weekday(), axis=1)
        print("                                                 UQDKR  l<  /.-~")
        dumptrack_2['etape'] = dumptrack_2.apply(lambda row: dico_etapes[(row.module,row.path,row.activity)],axis=1)
        print("                                                 ` l /~\ \<|Y")
        dumptrack_2['id_mpa'] = dumptrack_2.apply(lambda row: [int(row.module),int(row.path),int(row.activity)], axis=1)
        print("                                                    ',-~\ \L|")
        dumptrack_2['id_mpae'] = dumptrack_2.apply(lambda row: id_mpae_3_step(row), axis=1)
        print("'--.____  '--------.______       _.----.-----./      :/   '--'")
        dumptrack_2['date'] = dumptrack_2.apply(lambda row: (datetime.datetime.fromtimestamp(row.tms).year,datetime.datetime.fromtimestamp(row.tms).month,datetime.datetime.fromtimestamp(row.tms).day), axis=1)
        print("        '--.__            `'----/       '-.      __ :/")
        chatbot = set_essai(dumptrack_2)
        print("              '-.___           :           \   .'  )/")
        
        # we have to transform the id_mpae column in string :
        dfnew = chatbot[['id_mpae']].copy(deep=True)
        pds.options.mode.chained_assignment = None
        chatbot['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )
        chatbot['diagnostique'] = chatbot['id_mpae'].map(diagnostique)
        chatbot['id_mpae'] = dfnew['id_mpae'].apply( lambda x : list( int(e) for e in x ) )

        chatbot.loc[chatbot['etape']!=1,'diagnostique'] = np.nan
        print("                    '---._           _.-'   ] /  _/")
        chatbot_new = duree(chatbot)
        print("                         '-._      _/     _/ / _/")
        chatbot_new = chatbot_new.sort_values('tms', axis=0, ascending=True)
        chatbot_new = pds.DataFrame.set_index(chatbot_new,np.arange(chatbot_new.shape[0]))
        print("                             \_ .-'____.-'__< |  \___")
        connection_index = get_index_connection(chatbot_new)
        chatbot_new['num_connection'] = chatbot_new.index.map(connection_index)
        print("                               <_______.\    \_\_---.7")
        came_back = {}
        for idt in chatbot_new['token'].unique() :
            came_back[idt] = sum(chatbot_new[chatbot_new['token']==idt]['duree'] > 2700)
        print("                              |   /'=r_.-'     _\\ =/")
        chatbot_new['reconnections'] = chatbot_new['token'].map(came_back)
        print("                          .--'   /            ._/'>")
        nbr_questions = {}
        for idt in chatbot_new['token'].unique() :
            nbr_questions[idt] = chatbot_new[chatbot_new['token']==idt].shape[0]
        print("                        .'   _.-'")
        chatbot_new['parcours_diagnostique'] = chatbot_new['diagnostique'].apply(get_parcours_diagnostique)
        chatbot_new['nbr_questions_faites'] = chatbot_new['token'].map(nbr_questions)
        print("                       / .--'")
        horaire = tranche_horaire(chatbot_new)
        print("                      /,/")
        chatbot_new['horaire'] = chatbot_new.apply( lambda row: horaire[(row.token,row.num_connection)],axis=1 )
        chatbot_new.columns = ['id_eleve', 'module', 'path', 'activity', 'tps_posix', 'exercice', 'correct', 'question_type',
       'jour', 'etape', 'id_mpa', 'id_mpae', 'date', 'essai', 'diagnostique',
       'tps+1', 'duree', 'num_connection', 'reconnections',
       'parcours_diagnostique', 'nbr_questions_faites', 'horaire']
        print("                      |/`)")
        chatbot_final = chatbot_new.loc[:,['id_eleve', 'module', 'path', 'activity', 'tps_posix', 'exercice','correct', 'question_type',
       'jour', 'etape', 'id_mpa', 'id_mpae', 'date', 'essai',
       'tps+1', 'duree', 'num_connection', 'reconnections',
       'parcours_diagnostique', 'nbr_questions_faites', 'horaire']]
        print("                      'c=,")
        chatbot_final.to_pickle((os.path.join(".","data", f"chatbot_data_{today}.pk1")))
        print("\nchatbot dataframe built from trackings and sequences data")

    if option == 1: # save 1 pandas dataframe : sequences

        print("building sequences dataframe\n")
        to_workable_json(original_json_sequences_name)
        json_to_csv_seq('jsonfile_to_csv_sequences2',f'transformed_{original_json_sequences_name}')
        print("                                              ,","                                             ,o","                                             :o",sep="\n")
        dumpseq_1 = csv_to_dataframe('jsonfile_to_csv_sequences2',['real_module','real_path','real_activity','real_exercice','difficultyseq', 'type_answer','module','path','activity','id_eleve', 'reponse_eleve', 'reponse_correcte'])
        print("                    _....._                  `:o","                  .'       ``-.                \o","                 /  _      _   \                \o",sep="\n")
        dumpseq_1[['real_module','real_path','real_activity','real_exercice']] = dumpseq_1[['real_module','real_path','real_activity','real_exercice']].apply(pds.to_numeric, errors='coerce')
        print("                :  /*\    /*\  :                 ;o","                |  \_/    \_/  :                  ;o","                (       U      /                  ;o",sep="\n")
        dumpseq_1[['real_module','real_path','real_activity','real_exercice']] = dumpseq_1[['real_module','real_path','real_activity','real_exercice']].astype('Int16')
        print("                 \  (\_____/) /                  /o","                  \   UQDKR  (                  /o","                   \         (                ,o:",sep="\n")
        dumpseq_1 = dumpseq_1.fillna(-1)
        print("                   )          \,           .o;o'           ,o'o'o.","                 ./          /\o;o,,,,,;o;o;''         _,-o,-'''-o:o.","  .             ./o./)        \    'o'o'o''         _,-'o,o'         o",sep="\n")
        dumpseq_1.loc[:,'id_mpae'] = dumpseq_1.apply(lambda row: [int(row.module),int(row.path),int(row.activity),int(row.real_exercice)], axis=1)
        print("  o           ./o./ /       .o \.              __,-o o,o'","  \o.       ,/o /  /o/)     | o o'-..____,,-o'o o_o-'","  `o:o...-o,o-' ,o,/ |     \   'o.o_o_o_o,o--''",sep="\n")
        dumpseq_1.loc[:,'id_final_mpae'] = dumpseq_1.apply(lambda row: [int(row.real_module),int(row.real_path),int(row.real_activity),int(row.real_exercice)], axis=1)
        print("  .,  ``o-o'  ,.oo/   'o /\.o`.","  `o`o-....o'o,-'   /o /   \o \.                       ,o..         o","    ``o-o.o--      /o /      \o.o--..          ,,,o-o'o.--o:o:o,,..:o",sep="\n")
        diagnostique = dict_diagnostique(dumpseq_1)
        print("                  (oo(          `--o.o`o---o'o'o,o,-'''        o'o'o","                   \ o\              ``-o-o''''","    ,-o;o           \o \\","    ,-o;o           \o \\",sep="\n")
        
        # we have to transform the id_mpae column in string :
        dfnew = dumpseq_1[['id_mpae']].copy(deep=True)
        pds.options.mode.chained_assignment = None
        dumpseq_1['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )
        dumpseq_1['diagnostique'] = dumpseq_1['id_mpae'].map(diagnostique)
        dumpseq_1['id_mpae'] = dfnew['id_mpae'].apply( lambda x : list( int(e) for e in x ) )

        print("   /o/               )o )","  (o(               /o /","   \o\.       ...-o'o /",sep="\n")
        dumpseq_1.to_pickle((os.path.join(".","data", f"sequences_data_{today}.pk1")))
        print("    \ o`o`-o'o o,o,--'","       ```o--'''",sep="\n")
        print("\nsequences dataframe built")

    if option == 2: #save 1 pandas dataframe : trackings

        print("\nbuilding sequences data\n")
        to_workable_json(original_json_sequences_name)
        json_to_csv_seq('jsonfile_to_csv_sequences2',f'transformed_{original_json_sequences_name}')
        print("                                              ,","                                             ,o","                                             :o",sep="\n")
        dumpseq_1 = csv_to_dataframe('jsonfile_to_csv_sequences2',['real_module','real_path','real_activity','real_exercice','difficultyseq', 'type_answer', 'module','path','activity','id_eleve', 'reponse_eleve', 'reponse_correcte'])
        print("                    _....._                  `:o","                  .'       ``-.                \o","                 /  _      _   \                \o",sep="\n")
        dumpseq_1[['real_module','real_path','real_activity','real_exercice']] = dumpseq_1[['real_module','real_path','real_activity','real_exercice']].apply(pds.to_numeric, errors='coerce')
        print("                :  /*\    /*\  :                 ;o","                |  \_/    \_/  :                  ;o","                (       U      /                  ;o",sep="\n")
        dumpseq_1[['real_module','real_path','real_activity','real_exercice']] = dumpseq_1[['real_module','real_path','real_activity','real_exercice']].astype('Int16')
        print("                 \  (\_____/) /                  /o","                  \   UQDKR  (                  /o","                   \         (                ,o:",sep="\n")
        dumpseq_1 = dumpseq_1.fillna(-1)
        print("                   )          \,           .o;o'           ,o'o'o.","                 ./          /\o;o,,,,,;o;o;''         _,-o,-'''-o:o.","  .             ./o./)        \    'o'o'o''         _,-'o,o'         o",sep="\n")
        dumpseq_1.loc[:,'id_mpae'] = dumpseq_1.apply(lambda row: [int(row.module),int(row.path),int(row.activity),int(row.real_exercice)], axis=1)
        print("  o           ./o./ /       .o \.              __,-o o,o'","  \o.       ,/o /  /o/)     | o o'-..____,,-o'o o_o-'","  `o:o...-o,o-' ,o,/ |     \   'o.o_o_o_o,o--''",sep="\n")
        dumpseq_1.loc[:,'id_final_mpae'] = dumpseq_1.apply(lambda row: [int(row.real_module),int(row.real_path),int(row.real_activity),int(row.real_exercice)], axis=1)
        print("  .,  ``o-o'  ,.oo/   'o /\.o`.","  `o`o-....o'o,-'   /o /   \o \.                       ,o..         o","    ``o-o.o--      /o /      \o.o--..          ,,,o-o'o.--o:o:o,,..:o",sep="\n")
        diagnostique = dict_diagnostique(dumpseq_1)
        print("                  (oo(          `--o.o`o---o'o'o,o,-'''        o'o'o","                   \ o\              ``-o-o''''","    ,-o;o           \o \\","    ,-o;o           \o \\",sep="\n")
        
        # we have to transform the id_mpae column in string :
        dfnew = dumpseq_1[['id_mpae']].copy(deep=True)
        pds.options.mode.chained_assignment = None
        dumpseq_1['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )
        dumpseq_1['diagnostique'] = dumpseq_1['id_mpae'].map(diagnostique)
        dumpseq_1['id_mpae'] = dfnew['id_mpae'].apply( lambda x : list( int(e) for e in x ) )

        print("   /o/               )o )","  (o(               /o /","   \o\.       ...-o'o /",sep="\n")
        print("    \ o`o`-o'o o,o,--'","       ```o--'''",sep="\n")
        print("\nbuilding trackings data\n")
        to_workable_json(original_json_trackings_name)
        json_to_csv_track('jsonfile_to_csv_trackings2',f'transformed_{original_json_trackings_name}')
        print("                                                       ______","                                                     /    /_-._",sep="\n")
        print("                                                   / /_ ~~o\  :Y","                                                   / : \~x.  ` ')",sep="\n")
        dumptrack_2 = csv_to_dataframe('jsonfile_to_csv_trackings2',['token','module','path','activity','tms','exerciceId','isCorrect','question_type'])
        dico_etapes = remplir_etape()
        print("                                                  /  |  Y< ~-.__j")
        dumptrack_2['tms'] = dumptrack_2['tms']/1000
        dumptrack_2['jour'] = dumptrack_2.apply(lambda row: datetime.datetime.fromtimestamp(row.tms).weekday(), axis=1)
        print("                                                 UQDKR  l<  /.-~")
        dumptrack_2['etape'] = dumptrack_2.apply(lambda row: dico_etapes[(row.module,row.path,row.activity)],axis=1)
        print("                                                 ` l /~\ \<|Y")
        dumptrack_2['id_mpa'] = dumptrack_2.apply(lambda row: [int(row.module),int(row.path),int(row.activity)], axis=1)
        print("                                                    ',-~\ \L|")
        dumptrack_2['id_mpae'] = dumptrack_2.apply(lambda row: id_mpae_3_step(row), axis=1)
        print("'--.____  '--------.______       _.----.-----./      :/   '--'")
        dumptrack_2['date'] = dumptrack_2.apply(lambda row: (datetime.datetime.fromtimestamp(row.tms).year,datetime.datetime.fromtimestamp(row.tms).month,datetime.datetime.fromtimestamp(row.tms).day), axis=1)
        print("        '--.__            `'----/       '-.      __ :/")
        chatbot = set_essai(dumptrack_2)
        print("              '-.___           :           \   .'  )/")
        
        # we have to transform the id_mpae column in string :
        dfnew = chatbot[['id_mpae']].copy(deep=True)
        pds.options.mode.chained_assignment = None
        chatbot['id_mpae'] = dfnew['id_mpae'].apply( lambda x : ''.join(str(e) for e in x) )
        chatbot['diagnostique'] = chatbot['id_mpae'].map(diagnostique)
        chatbot['id_mpae'] = dfnew['id_mpae'].apply( lambda x : list( int(e) for e in x ) )

        chatbot.loc[chatbot['etape']!=1,'diagnostique'] = np.nan
        print("                    '---._           _.-'   ] /  _/")
        chatbot_new = duree(chatbot)
        print("                         '-._      _/     _/ / _/")
        chatbot_new = chatbot_new.sort_values('tms', axis=0, ascending=True)
        chatbot_new = pds.DataFrame.set_index(chatbot_new,np.arange(chatbot_new.shape[0]))
        print("                             \_ .-'____.-'__< |  \___")
        connection_index = get_index_connection(chatbot_new)
        chatbot_new['num_connection'] = chatbot_new.index.map(connection_index)
        print("                               <_______.\    \_\_---.7")
        came_back = {}
        for idt in chatbot_new['token'].unique() :
            came_back[idt] = sum(chatbot_new[chatbot_new['token']==idt]['duree'] > 2700)
        print("                              |   /'=r_.-'     _\\ =/")
        chatbot_new['reconnections'] = chatbot_new['token'].map(came_back)
        print("                          .--'   /            ._/'>")
        nbr_questions = {}
        for idt in chatbot_new['token'].unique() :
            nbr_questions[idt] = chatbot_new[chatbot_new['token']==idt].shape[0]
        print("                        .'   _.-'")
        chatbot_new['parcours_diagnostique'] = chatbot_new['diagnostique'].apply(get_parcours_diagnostique)
        chatbot_new['nbr_questions_faites'] = chatbot_new['token'].map(nbr_questions)
        print("                       / .--'")
        horaire = tranche_horaire(chatbot_new)
        print("                      /,/")
        chatbot_new['horaire'] = chatbot_new.apply( lambda row: horaire[(row.token,row.num_connection)],axis=1 )
        # WARNING : Pay attention of the current order of columns here befor changing the names :
        # print(chatbot_new.columns)
        chatbot_new.columns = ['id_eleve', 'module', 'path', 'activity', 'tps_posix', 'exercice', 'correct', 'question_type',
       'jour', 'etape', 'id_mpa', 'id_mpae', 'date', 'essai', 'diagnostique',
       'tps+1', 'duree', 'num_connection', 'reconnections', 
       'parcours_diagnostique', 'nbr_questions_faites', 'horaire']
        print("                      |/`)")
        # Now we can delete some unuseful columns 
        chatbot_final = chatbot_new.loc[:,['id_eleve', 'module', 'path', 'activity', 'tps_posix', 'exercice','correct', 'question_type',
       'jour', 'etape', 'id_mpa', 'id_mpae', 'date', 'essai',
       'tps+1', 'duree', 'num_connection', 'reconnections',
       'parcours_diagnostique', 'nbr_questions_faites', 'horaire']]
        print("                      'c=,")
        chatbot_final.to_pickle((os.path.join(".","data", f"chatbot_data_{today}.pk1")))
        print("\nchatbot dataframe built from trackings and sequences data\n")

#transform_data(3, 'sequences_original_v2','trackings_original_v2')

def main(argv):
    if usage(argv) != 0:
        exit(0)
    if check_error(argv) != 0:
        exit(84)
    option = int(argv[1])
    seq_name = str(argv[2])
    track_name = str(argv[3])
    transform_data(option, seq_name, track_name)
    return 0

if __name__ == '__main__':
    main(sys.argv)




## EVIDENCEB, 2020
## library of function to use
## File description:
## functions to use in order to avoid pandas problem working with list

import pandas as pds

def str_question_tolist(question):
    to_return = list( int(e) for e in question )
    if len(to_return)==5 :
        new_return = to_return[:3]
        new_return.append(int( str(to_return[3])+str(to_return[4]) ))
        return new_return
    else :
        return to_return

def list_question_tostr(question):
    return ''.join(str(e) for e in question)

def from_list_to_str(df,column_name):
    # to avoid pandas error from manipulating lists in dataframe we convert question into string
    dfnew = df[[f"{column_name}"]].copy(deep=True)
    pds.options.mode.chained_assignment = None
    df[f"{column_name}"] = dfnew[f"{column_name}"].apply( lambda x : list_question_tostr(x) )
    return df

def from_str_to_list(df,column_name):
    # to get a better visualisation of questions at the end of the algorithm
    dfnew = df[[f"{column_name}"]].copy(deep=True)
    pds.options.mode.chained_assignment = None
    df[f"{column_name}"] = dfnew[f"{column_name}"].apply( lambda x : str_question_tolist(x) )
    return df




## EVIDENCEB, 2020
## dropout.py
## File description:
## dropout function

import sys
from LrnXPAnaToolbox.lib import *
import pandas as pds
import numpy as np
import os
import datetime
from tqdm import tqdm
from tqdm.auto import tqdm
from sklearn.model_selection import train_test_split, GridSearchCV
from sklearn.tree import DecisionTreeClassifier

def usage(argv): # argv = arguments passed to the script = dropout.py -h
    if len(argv) == 2:
        if argv[1] == "-h":
            print("\nUSAGE")
            print("    ./dropout.py  new_student_data  available_database  recommendation_dataset")
            print("\nDESCRIPTION")
            print("""    new_student_data       Dataframe with all the data for only one (new) student (created with transform_data.py for example)""")
            print("""    available_database       Dataframe with all the available data (created with transform_data.py for example)""")
            print("""    recommendation_dataset       Dataframe with the N best questions to propose to the new_student""")
            return 1
    return 0

def check_error(argv): # argv = arguments passed to the script = dropout.py new_student_data available_database recommendation_dataset
    if len(argv) != 4:
        print("Wrong number of arguments, try \"-h\" for more informations")
        return 1
    if (argv[1].isidentifier() != True)|(argv[2].isidentifier() != True)|(argv[3].isidentifier() != True):
        print("new_student_data, available_database and recommendation_dataset arguments have to be the name of pandas dataframe so : alphanumeric letters (a-z) and/or (0-9) and/or underscores (_).")
        return 1
    return 0

def dropout_prediction_training_data(dataset, module_concerned, path_concerned, dropout_after_activity):
    """Compute students features to train the dropout prediciton on."""
    module_concerned = int(module_concerned)
    path_concerned = int(path_concerned)
    dropout_after_activity = int(dropout_after_activity)
    d = {}
    grouped = dataset[(dataset.module==module_concerned)&(dataset.path==path_concerned)]
    filtered = grouped.groupby('id_eleve')
    for student, student_groupby in filtered :
        if student_groupby[student_groupby.activity == dropout_after_activity].shape[0] != 0 :
            # the number of the first connection in which the student has done the activity dropout_after_activity
            first_connection_for_dropout_after_activity = student_groupby[student_groupby.activity == dropout_after_activity].sort_values('tps_posix', ascending=True).iloc[0]['num_connection']
            # the date of the last exercise the student has done in the activity dropout_after_activity
            highest_date_dropout_after_activity = max(student_groupby[ (student_groupby.activity == dropout_after_activity) & (student_groupby.num_connection == first_connection_for_dropout_after_activity) ]['tps_posix'])
            student_groupby = student_groupby[ (student_groupby.num_connection == first_connection_for_dropout_after_activity) & (student_groupby.tps_posix <= highest_date_dropout_after_activity) ]
            if student_groupby.shape[0] != 0 :
                d[student] = [student]
                # 1st feature : average_response_time
                average_response_time = sum(student_groupby['duree']) / student_groupby.shape[0]
                d[student].append( average_response_time )
                # 2nd feature : success_rate
                success_rate = student_groupby[student_groupby.correct == True].shape[0] / student_groupby.shape[0]
                d[student].append( success_rate )
                # 3rd feature : first assigned_path after the diagnostic test
                if student_groupby[ student_groupby.etape == 2 ].shape[0] == 0:
                    path = 0
                else :
                    path = student_groupby[ student_groupby.etape == 2 ].sort_values('tps_posix',ascending=True).iloc[0]['path']
                d[student].append( path )
                # 4th feature : the day of the fisrt connection in which the student has done the activity
                day = student_groupby.iloc[0]['jour']
                d[student].append( day )
                # 5th feature : the timeslot of the connection
                timeslot = student_groupby.iloc[0]['horaire']
                d[student].append( timeslot )
                # last feature :
                # 0 if the student dropout
                # 1 if the student will reach the next activity : the activity after the activity dropout_after_activity
                if student_groupby[ student_groupby.activity == dropout_after_activity + 1 ].shape[0] == 0 :
                    d[student].append( 0 )
                else :
                    d[student].append( 1 )
    X = np.array(list(d.values()))
    DF = pds.DataFrame(X)
    DF.columns = ['student','average_response_time', 'success_rate', 'assigned_path', 'day', 'timeslot', 'dropout']
    return DF

def dropout_prediction(prediction_test, student_array_features):
    """Return the dropout prediciton value (0 if dropout, 1 if not) and the confidence level of the prediciton.
       The non linear classification method used is the Decision Tree Classifier from sklearn. """
    X = prediction_test.drop('dropout', axis = 1)
    Y = prediction_test['dropout']
    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size = 0.3)
    parametre = [{"max_depth":list(range(2,10))}]
    retry = True
    while retry = True :
        try :
            try_model = GridSearchCV(DecisionTreeClassifier(),parametre,n_jobs=-1)
            try_model.fit(X_train,Y_train)
            retry = False
        except ValueError :
            retry = True
    best_param = try_model.fit(X_train,Y_train).best_params_
    model = DecisionTreeClassifier(max_depth = best_param["max_depth"])
    model.fit(X_train, Y_train)
    return model.predict( student_array_features ), model.score( X_test, Y_test )

def dropout_recommendation(new_student_data, available_database, recommendation_dataset):
    """ parameters :
        - new_student_data  : student for whom we produced 20 best questions
          (there must be the columns : student, duree, correct, etape, jour, horaire)
        - available_database of all the available data : example : chatbot_data.pk1,
          there must be all the data collected but we suppose that the new data of our student are not yet in.
          If there is a way to put them in directly : no need for the parameters new_student_data because we can compute
          this parameters thanks to the parameter : available_database
        - recommendation_dataset of the 20 best questions to propose, columns  ['student_id','question_id','rating'].
        returns :
        - the recommendation_dataset modified according to the predicted dropout prediciton value."""
    # if the available_database contains the student data :
    # we can get the following data for student_array_features with the available_database
    # else : we get student information thanks to the new_student_data parameter
    student = new_student_data['id_eleve']
    student_array_features = np.array([])
    # 1st feature : average_response_time
    average_response_time = sum(new_student_data['duree']) / new_student_data.shape[0]
    student_array_features = np.append( student_array_features, average_response_time )
    # 2nd feature : success_rate
    success_rate = new_student_data[new_student_data.correct == True].shape[0] / new_student_data.shape[0]
    student_array_features = np.append( student_array_features, success_rate )
    # 3rd feature : first assigned_path after the diagnostic test
    if new_student_data[ new_student_data.etape == 2 ].shape[0] == 0:
        path = 0
    else :
        path = new_student_data[ new_student_data.etape == 2 ].sort_values('tps_posix',ascending=True).iloc[0]['path']
    student_array_features = np.append( student_array_features, path )
    # 4th feature : the day of the fisrt connection in which the student has done the activity
    day = new_student_data.iloc[0]['jour']
    student_array_features = np.append( student_array_features, day )
    # 5th feature : the timeslot of the connection
    timeslot = new_student_data.iloc[0]['horaire']
    student_array_features = np.append( student_array_features, timeslot )
    dropout_index = np.array([])
    student_array_features = np.array([student_array_features])
    # question in recommendation_dataset are supposed to be list of the form [module, parcours, activite, exo]

    # to avoid pandas error from manipulating lists in dataframe we convert question into string
    from_list_to_str(recommendation_dataset,'question_id')

    print(f"WARNING : if UserWarning from sklearn\model_selection\_split : not enough participants for the question.")
    for question_id in tqdm(recommendation_dataset['question_id']) :
        prediction_test = dropout_prediction_training_data(available_database, module_concerned=question_id[0] , path_concerned=question_id[1], dropout_after_activity=question_id[2]).drop(['student'], axis = 1)
        # student_dropout = 0 : the student will probably dropout
        # student_dropout = 1 : the student will probably continue
        student_dropout, confidence_rating = dropout_prediction(prediction_test, student_array_features)
        if confidence_rating > 0.86 :
            if student_dropout[0] == str(0) :
                dropout_index = np.append( dropout_index, int(recommendation_dataset[recommendation_dataset['question_id']==question_id].index[0]) )

    from_str_to_list(recommendation_dataset,'question_id')
    new_recommendation_dataset = recommendation_dataset.drop(dropout_index, axis=0) # WARNING : new_recommendation_dataset could be an empty dataframe
    return new_recommendation_dataset

def main(argv):
    if usage(argv) != 0:
        exit(0)
    if check_error(argv) != 0:
        exit(84)
    new_student_data = argv[1]
    available_database = argv[2]
    recommendation_dataset = argv[3]
    dropout_recommendation(new_student_data, available_database, recommendation_dataset)
    return 0

if __name__ == '__main__':
    main(sys.argv)



## EVIDENCEB, 2020
## cluster_students.py
## File description:
## cluster students function

import pandas as pds
import numpy as np
import os
import datetime
from matplotlib import pyplot as plt
from sklearn.cluster import KMeans, MiniBatchKMeans, AgglomerativeClustering
from sklearn.metrics import pairwise_distances_argmin_min
from sklearn import metrics
from sklearn import preprocessing
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
import sys

def usage(argv): # argv = arguments passed to the script = cluster_students.py -h
    if len(argv) == 2:
        if argv[1] == "-h":
            print("\nUSAGE")
            print("    ./cluster_students.py  available_database  new_student_data  clustering_plot")
            print("\nDESCRIPTION")
            print("""    available_database       dataframe with all the data (created with transform_data.py for example)""")
            print("""    new_student_data       dataframe with all the data for only one (new) student (created with transform_data.py for example)""")
            print("""    clustering_plot       default value set to False, if True the function plot the 3D PCA graph of the clustering""")
            return 1
    return 0

def check_error(argv): # argv = arguments passed to the script = cluster_students.py available_database new_student_data clustering_plot
    if (len(argv) != 4) & (len(argv) != 3) :
        print("Wrong number of arguments, try \"-h\" for more informations")
        return 1
    if len(argv) == 4 :
        if argv[3] not in ['True','False'] :
            print("Argument have to be True if you want to see the graphic representation of the clustering or False if not.")
            return 1
        else :
            return 0
    elif len(argv) != 3 :
        return 1
    else :
        return 0
    #if (len(argv) != 4) & (len(argv) != 3) :
    #    print("Wrong number of arguments, try \"-h\" for more informations")
    #    return 1
    #if len(argv) == 4 :
    #    if argv[3] not in ['True','False'] :
    #        print("Argument have to be True if you want to see the graphic representation of the clustering or False if not.")
    #        return 1
    #    if (argv[2].isidentifier() != True)|(argv[1].isidentifier() != True) :
    #        print("available_database and new_student_data arguments have to be the name of pandas dataframe so : alphanumeric letters (a-z) and/or (0-9) and/or underscores (_).")
    #        return 1
    #elif len(argv) == 3 :
    #    if (argv[2].isidentifier() != True)|(argv[1].isidentifier() != True) :
    #        print("available_database and new_student_data arguments have to be the name of pandas dataframe so : alphanumeric letters (a-z) and/or (0-9) and/or underscores (_).")
    #        return 1
    #else :
    #    return 1
    #return 0

def plot_clustering(DFVariables, DFMBKMEANS, nbr_clusters, nbr_components=3) :
    """Show and save in the data folder the figure of a clustering given with the DFMBKMEANS table."""
    # standardisation
    featuresPCA = DFVariables.copy()
    featuresPCA = StandardScaler().fit_transform(featuresPCA)
    pca = PCA(svd_solver='full')
    pca.fit_transform(featuresPCA)
    print(f"\nWith 3 principal components we explain {round((pca.explained_variance_ratio_[0]+pca.explained_variance_ratio_[1]+pca.explained_variance_ratio_[2])*100,2)}% of the total information,\nthe default number of principal components is set to 3.")

    n = DFVariables.shape[0] # observations
    p = DFVariables.shape[1] # variables

    # SCREE PLOT
    # corrected explained variance
    # var = (n-1)/n*pca.explained_variance_
    # threshold for the scree plot
    # bs = 1/np.arange(p,0,-1)
    # bs = np.cumsum(bs)
    # bs = bs[::-1]
    # scree plot values : we want eigenvalue > threshold
    # print(pds.DataFrame({'eigenvalue':var,'threshold':bs}))

    # 3D PCA Projection :
    pca = PCA(n_components=nbr_components)
    principalComponents = pca.fit_transform(featuresPCA)
    principalDf = pds.DataFrame(data = principalComponents, columns = ['principal component 1', 'principal component 2', 'principal component 3'])
    finalDf = pds.concat([principalDf, DFMBKMEANS[['Cluster']]], axis = 1)
    fig = plt.figure(figsize = (10,10))
    ax = fig.add_subplot(111, projection='3d')
    ax.view_init(25,-120)
    xAxisLine = ((min(principalComponents[:,0]), max(principalComponents[:,0])), (0, 0), (0,0))
    yAxisLine = ((0, 0), (min(principalComponents[:,1]), max(principalComponents[:,1])), (0,0))
    zAxisLine = ((0, 0), (0,0), (min(principalComponents[:,2]), max(principalComponents[:,2])))
    ax.set_xlabel('Principal Component 1', fontsize = 15)
    ax.set_ylabel('Principal Component 2', fontsize = 15)
    ax.set_zlabel('Principal Component 3', fontsize = 15)
    ax.set_title('3 Components PCA of the clustering', fontsize = 20)
    targets = np.arange(nbr_clusters)
    colors_to_chose = ['peachpuff','lightseagreen','peru','b','g','r', 'c','m', 'y', 'k','sandybrown','seashell','turquoise','aquamarine','lightcyan']
    colors = colors_to_chose[:nbr_clusters]
    for target, color in zip(targets,colors):
        indicesToKeep = finalDf['Cluster'] == target
        ax.scatter(finalDf.loc[indicesToKeep, 'principal component 1']
                   , finalDf.loc[indicesToKeep, 'principal component 2']
                   , finalDf.loc[indicesToKeep, 'principal component 3']
                   , c = color
                   , s = 50)
    ax.legend(targets)
    ax.grid()
    plt.show()
    today = datetime.date.today()
    plt.savefig( (os.path.join(".","data", f"MBKMeans_figure_{today}.png")) )

def creationDFClustering(available_database):
    """Create the individuals representation vectors with some features."""
    d = {}
    grouped = available_database[(available_database.num_connection==1)]
    filtered = grouped.groupby('id_eleve')
    for student, student_groupby in filtered :

        # we first add the student
        d[student] = [student]

        student_data = available_database[available_database.id_eleve == student]

        # we add the average response time
        d[student].append( sum(student_data.duree)/(student_data.shape[0]) )

        # we add the success rate
        d[student].append( student_data[student_data.correct == True].shape[0]/student_data.shape[0] )

        # we add the path assigned to the student
        path_data = student_data[student_data.etape == 2]
        if path_data.shape[0] == 0 :
            d[student].append(0)
        else:
            path_data.sort_values(by = 'tps_posix', ascending = True)
            d[student].append( path_data.iloc[0]['path'] )

        # we add the total time spent on the chatbot
        d[student].append( sum(student_data.duree) )

        # we add the number of questions the student tried
        d[student].append( student_data.shape[0] )

        # we add the day
        day = student_data.iloc[0]['jour']
        d[student].append( day )

        # we add the timeslot
        timeslot = student_data.iloc[0]['horaire']
        d[student].append( timeslot )

    df = pds.DataFrame(np.array(list(d.values())))
    df.columns = ['id_eleve','temps_moyen_reponse','taux_reussite','parcours','duree_totale','nbr_questions','jour','horaire']
    return df

def optimal_n_clusters(data, clustering_method, nbr_students):
    """ Choose the optimal number (between 3 and 30) of cluster to use according to some criteria :
        - the minimal population of the clustering (so the number of individuals in the smallest cluster) must be >= to 1/50*the total number of individuals
          for example if there are 2500 students in the data, the minimal population must be >= 50 individuals
        - the maximal population of the clustering (so the number of individuals in the biggest cluster) must be <= to 1/2.5*the total number of individuals
          for example if there are 2500 students in the data, the maximal population must be <= 1000 individuals
    """
    optimal = False
    while optimal == False :
        choice = np.array([])
        population_min = np.array([])
        for k in range(28): # k : 0+3 > 27+3
            cls = clustering_method(n_clusters=k+3)
            cls.fit(data)
            groups = cls.labels_
            unique, counts = np.unique(groups, return_counts=True)
            dic = dict(zip(unique, counts))
            if (min(dic.values()) >= 1/50*nbr_students) & (max(dic.values()) <= 1/2.5*nbr_students) :
                choice = np.append(choice, metrics.silhouette_score(data, groups))
            else :
                choice = np.append(choice, -1)
        if np.mean( choice ) != -1 :
            optimal = True
    return np.flip(np.argsort(choice)[-3:])+3

def similar_students(available_database, new_student_data, clustering_plot=False) :
    """ Return 2 elements :
        - a table with 2 columns : students, clusters
        - the predicted cluster for an array of data (array of the data for one student which is not supposed to be in the data).
    
        Save the MiniBatchKMeans dataframe to pickle in the data folder
    """
    student_array_features = np.array([])
    # we add the average response time
    student_array_features = np.append( student_array_features, sum(new_student_data.duree)/(new_student_data.shape[0]) )
    # we add the success rate
    student_array_features = np.append( student_array_features,new_student_data[new_student_data.correct == True].shape[0]/new_student_data.shape[0] )
    # we add the path assigned to the student
    # WARNING : if the student has done 2 modules at his/her first connection we'll have 2 assigned path, which one to choose?
    path_data = new_student_data[new_student_data.etape == 2]
    if path_data.shape[0] == 0 :
        student_array_features = np.append( student_array_features,0 )
    else:
        path_data.sort_values(by = 'tps_posix', ascending = True)
        student_array_features = np.append( student_array_features,path_data.iloc[0]['path'] )
    # we add the total time spent on the chatbot
    student_array_features = np.append( student_array_features,sum(new_student_data.duree) )
    # we add the number of questions the student tried
    student_array_features = np.append( student_array_features,new_student_data.shape[0] )
    # we add the day
    day = new_student_data.iloc[0]['jour']
    student_array_features = np.append( student_array_features,day )
    # we add the timeslot
    timeslot = new_student_data.iloc[0]['horaire']
    student_array_features = np.append( student_array_features,timeslot )
    student_array_features = np.array([student_array_features])

    clustering_method = MiniBatchKMeans # the MiniBatchKMeans is the best method for our data.
    X = creationDFClustering(available_database)
    nbr_students = len(X['id_eleve'].unique())
    DFVariables = X.drop(['id_eleve'], axis = 1)
    DFVariables = DFVariables.apply(pds.to_numeric)
    normalized_data = preprocessing.normalize(DFVariables)
    nbr_clusters = optimal_n_clusters(normalized_data, clustering_method, nbr_students)[0]
    clustering_model = clustering_method(n_clusters=nbr_clusters)
    clustering = clustering_model.fit(normalized_data)
    DFMBKMEANS = X.copy()
    DFMBKMEANS['Cluster'] = clustering.labels_
    today = datetime.date.today()
    DFMBKMEANS.to_pickle((os.path.join(".","data", f"DF_MBKMeans_{today}.pk1")))
    if str(clustering_plot) == 'True' :
        print('\nPlease close the figure if ou want to continue.\n')
        plot_clustering(DFVariables, DFMBKMEANS, nbr_clusters)
    return DFMBKMEANS.loc[:,['id_eleve','Cluster']], clustering_model.predict(student_array_features)

def main(argv):
    if usage(argv) != 0:
        exit(0)
    if check_error(argv) != 0:
        exit(84)
    available_database = argv[1]
    new_student_data = argv[2]
    if len(argv)==4:
        clustering_plot = argv[3]
        similar_students(available_database, new_student_data, clustering_plot)
    elif len(argv)==3:
        similar_students(available_database, new_student_data)
    else:
        print("Wrong number of arguments, try \"-h\" for more informations")
    return 0

if __name__ == '__main__':
    main(sys.argv)


#!/usr/bin/python3

from LrnXPAnaToolbox.recommender.final_recommendation import recom_algorithm
from LrnXPAnaToolbox.recommender.dropout import dropout_recommendation
from LrnXPAnaToolbox.cluster_students.cluster_students import similar_students
import pandas as pds
import numpy as np
import os
import sys

def usage(argv):
    if len(argv) == 2:
        if argv[1] == "-h":
            print("\nUSAGE")
            print("    .\recommender.py  student_df  student_marks  all_students_df  all_student_marks")
            print("\nDESCRIPTION")
            print("    student_df           dataframe containing the dataframe of data of the student you")
            print("                         want to recommend questions")
            print("    student_marks        dataframe containing the dataframe of marks of the student for recommendations")
            print("    all_students_df      dataframe containing the dataframe of data of all students")
            print("    all_students_marks   dataframe containing the dataframe of marks of all students")
            print("\nAll files must be in a directory named \'data\'")
            return 1
    return 0

def check_error(argv):
    if len(argv) != 5:
        print("There must be 4 files in argument. Try -h for more information")
        return 1
    if os.path.isdir('data') != True:
        print("File must be in \"data\" directory. Try -h for more informations.")
        return 1
    for i in range(1, len(argv)):
        file = argv[i]
        path = os.path.join(".", "data", file)
        if os.path.isfile(path) != True:
            print("No such file or directory :", file)
            return 1
    return 0

def algorithm(student_data, student_marks, students_df, marks_df):
    """Compute the recommendation algorithm, check the dropout prediciton using only student from the same cluster 
       in order to optimize execution, select the questions for which dropout is not predicted (unless dropout is predicted for all question,
       in this case we just present all the questions without selection) and recommend these questions."""
    cluster_df, student_cluster = similar_students(students_df, student_data, False)
    cluster_df.columns = ['student_id', 'cluster_id']
    student_cluster = int(student_cluster[0])
    new_clust_df = cluster_df[cluster_df.cluster_id == student_cluster]
    students_df = students_df[students_df['id_eleve'].isin(new_clust_df['student_id'])]
    marks_df = marks_df[marks_df['student_id'].isin(new_clust_df['student_id'])]
    recom_quests = recom_algorithm(student_marks, marks_df)
    quest_after_dropout = dropout_recommendation(student_data, students_df, recom_quests)
    if quest_after_dropout.empty:
        quest_after_dropout = recom_quests
    print(f"\n\n\nRECOMMENDATION QUESTION(S) FOR {student_data.iloc[0].id_eleve} : \n\n{quest_after_dropout}")
    return quest_after_dropout

def get_files(argv):
    df1 = pds.read_pickle((os.path.join(".","data",argv[1])))
    df2 = pds.read_pickle((os.path.join(".","data",argv[2])))
    df3 = pds.read_pickle((os.path.join(".","data",argv[3])))
    df4 = pds.read_pickle((os.path.join(".","data",argv[4])))
    df2.columns = ['student_id', 'question_id', 'rating']
    df4.columns = ['student_id', 'question_id', 'rating']
    return df1, df2, df3, df4

def main(argv):
    if usage(argv) != 0:
        return 0
    if check_error(argv) != 0:
        return 84
    student_data, student_marks, students_df, marks_df = get_files(argv)
    recommendation = algorithm(student_data, student_marks, students_df, marks_df)
    pds.set_option('display.max_columns', None)
    return 0

if __name__ == '__main__':
    main(sys.argv)